#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import logging
import os
import platform
import stat
from datetime import timedelta

import mock
import pytest

import elasticapm.conf
from elasticapm.conf import (
    Config,
    ConfigurationError,
    EnumerationValidator,
    ExcludeRangeValidator,
    FileIsReadableValidator,
    PrecisionValidator,
    RegexValidator,
    SupportedValueInFipsModeValidator,
    UnitValidator,
    VersionedConfig,
    _BoolConfigValue,
    _ConfigBase,
    _ConfigValue,
    _DictConfigValue,
    _DurationConfigValue,
    _ListConfigValue,
    constants,
    setup_logging,
    size_validator,
)


def test_basic_not_configured():
    with mock.patch("logging.getLogger", spec=logging.getLogger) as getLogger:
        logger = getLogger()
        logger.handlers = []
        handler = mock.Mock()
        result = setup_logging(handler)
        assert result


def test_basic_already_configured():
    with mock.patch("logging.getLogger", spec=logging.getLogger) as getLogger:
        handler = mock.Mock()
        logger = getLogger()
        logger.handlers = [handler]
        result = setup_logging(handler)
        assert not result


def test_config_dict():
    config = Config(
        {
            "SERVICE_NAME": "foo",
            "SECRET_TOKEN": "bar",
            "SERVER_URL": "http://example.com:1234",
            "SERVICE_VERSION": 1,
            "HOSTNAME": "localhost",
            "API_REQUEST_TIME": "5s",
        }
    )

    assert config.service_name == "foo"
    assert config.secret_token == "bar"
    assert config.server_url == "http://example.com:1234"
    assert config.service_version == "1"
    assert config.hostname == "localhost"
    assert config.api_request_time.total_seconds() == 5


def test_config_environment():
    with mock.patch.dict(
        "os.environ",
        {
            "ELASTIC_APM_SERVICE_NAME": "foo",
            "ELASTIC_APM_SECRET_TOKEN": "bar",
            "ELASTIC_APM_SERVER_URL": "http://example.com:1234",
            "ELASTIC_APM_SERVICE_VERSION": "1",
            "ELASTIC_APM_HOSTNAME": "localhost",
            "ELASTIC_APM_API_REQUEST_TIME": "5s",
            "ELASTIC_APM_AUTO_LOG_STACKS": "false",
        },
    ):
        config = Config()

        assert config.service_name == "foo"
        assert config.secret_token == "bar"
        assert config.server_url == "http://example.com:1234"
        assert config.service_version == "1"
        assert config.hostname == "localhost"
        assert config.api_request_time.total_seconds() == 5
        assert config.auto_log_stacks is False


def test_config_inline_dict():
    config = Config(
        inline_dict={
            "service_name": "foo",
            "secret_token": "bar",
            "server_url": "http://example.com:1234",
            "service_version": "1",
            "api_request_time": "5s",
        }
    )

    assert config.service_name == "foo"
    assert config.secret_token == "bar"
    assert config.server_url == "http://example.com:1234"
    assert config.service_version == "1"
    assert config.hostname is None
    assert config.api_request_time.total_seconds() == 5


def test_config_precedence():
    #  precedence order: environment, inline dict, config dict
    with mock.patch.dict("os.environ", {"ELASTIC_APM_SERVICE_NAME": "bar"}):
        config = Config(
            {"SERVICE_NAME": "foo", "SECRET_TOKEN": "secret", "COLLECT_LOCAL_VARIABLES": "all"},
            inline_dict={"secret_token": "notsecret", "service_name": "baz"},
        )

    assert config.service_name == "bar"
    assert config.secret_token == "notsecret"
    assert config.collect_local_variables == "all"


def test_list_config_value():
    class MyConfig(_ConfigBase):
        my_list = _ListConfigValue("MY_LIST", list_separator="|", type=int)

    config = MyConfig({"MY_LIST": "1|2|3"})
    assert config.my_list == [1, 2, 3]


def test_dict_config_value():
    class MyConfig(_ConfigBase):
        my_dict = _DictConfigValue("MY_DICT")
        my_typed_dict = _DictConfigValue("MY_TYPED_DICT", type=int)
        my_native_dict = _DictConfigValue("MY_NATIVE_DICT")

    config = MyConfig(
        {"MY_DICT": "a=b, c = d ,e=f", "MY_TYPED_DICT": "a=1,b=2"}, inline_dict={"my_native_dict": {"x": "y"}}
    )
    assert config.my_dict == {"a": "b", "c": "d", "e": "f"}
    assert config.my_typed_dict == {"a": 1, "b": 2}
    assert config.my_native_dict == {"x": "y"}


def test_bool_config_value():
    class MyConfig(_ConfigBase):
        my_bool = _BoolConfigValue("MY_BOOL", true_string="yup", false_string="nope")

    config = MyConfig({"MY_BOOL": "yup"})

    assert config.my_bool is True

    config.my_bool = "nope"

    assert config.my_bool is False


def test_values_not_shared_among_instances():
    class MyConfig(_ConfigBase):
        my_bool = _BoolConfigValue("MY_BOOL", true_string="yup", false_string="nope")

    c1 = MyConfig({"MY_BOOL": "yup"})
    c2 = MyConfig({"MY_BOOL": "nope"})

    assert c1.my_bool is not c2.my_bool


def test_regex_validation():
    class MyConfig(_ConfigBase):
        my_regex = _ConfigValue("MY_REGEX", validators=[RegexValidator(r"\d+")])

    c1 = MyConfig({"MY_REGEX": "123"})
    c2 = MyConfig({"MY_REGEX": "abc"})
    assert not c1.errors
    assert "MY_REGEX" in c2.errors


def test_size_validation():
    class MyConfig(_ConfigBase):
        byte = _ConfigValue("BYTE", type=int, validators=[size_validator])
        kbyte = _ConfigValue("KBYTE", type=int, validators=[size_validator])
        mbyte = _ConfigValue("MBYTE", type=int, validators=[size_validator])
        gbyte = _ConfigValue("GBYTE", type=int, validators=[size_validator])
        wrong_pattern = _ConfigValue("WRONG_PATTERN", type=int, validators=[size_validator])

    c = MyConfig({"BYTE": "10b", "KBYTE": "5kb", "MBYTE": "17mb", "GBYTE": "2gb", "WRONG_PATTERN": "5 kb"})
    assert c.byte == 10
    assert c.kbyte == 5 * 1024
    assert c.mbyte == 17 * 1024 * 1024
    assert c.gbyte == 2 * 1024 * 1024 * 1024
    assert c.wrong_pattern is None
    assert "WRONG_PATTERN" in c.errors


def test_duration_validation():
    class MyConfig(_ConfigBase):
        microsecond = _DurationConfigValue("US", allow_microseconds=True)
        millisecond = _DurationConfigValue("MS")
        second = _DurationConfigValue("S")
        minute = _DurationConfigValue("M")
        default_unit_ms = _DurationConfigValue("DM", unitless_factor=0.001)
        wrong_pattern = _DurationConfigValue("WRONG_PATTERN")

    c = MyConfig({"US": "10us", "MS": "-10ms", "S": "5s", "M": "17m", "DM": "83", "WRONG_PATTERN": "5 ms"})
    assert c.microsecond == timedelta(microseconds=10)
    assert c.millisecond == timedelta(milliseconds=-10)
    assert c.second == timedelta(seconds=5)
    assert c.minute == timedelta(minutes=17)
    assert c.default_unit_ms == timedelta(milliseconds=83)
    assert c.wrong_pattern is None
    assert "WRONG_PATTERN" in c.errors


def test_precision_validation():
    class MyConfig(_ConfigBase):
        sample_rate = _ConfigValue("SR", type=float, validators=[PrecisionValidator(4, 0.0001)])

    c = MyConfig({"SR": "0.0000001"})
    assert c.sample_rate == 0.0001
    c = MyConfig({"SR": "0.555555"})
    assert c.sample_rate == 0.5556


def test_chained_validators():
    class MyConfig(_ConfigBase):
        chained = _ConfigValue("CHAIN", validators=[lambda val, field: val.upper(), lambda val, field: val * 2])

    c = MyConfig({"CHAIN": "x"})
    assert c.chained == "XX"


def test_file_is_readable_validator_not_exists(tmpdir):
    validator = FileIsReadableValidator()
    with pytest.raises(ConfigurationError) as e:
        validator(tmpdir.join("doesnotexist").strpath, "path")
    assert "does not exist" in e.value.args[0]


def test_file_is_readable_validator_not_a_file(tmpdir):
    validator = FileIsReadableValidator()
    with pytest.raises(ConfigurationError) as e:
        validator(tmpdir.strpath, "path")
    assert "is not a file" in e.value.args[0]


@pytest.mark.skipif(
    platform.system() == "Windows" or os.getuid() == 0,
    reason="os.access() doesn't seem to work as we expect on Windows and test will fail as root user",
)
def test_file_is_readable_validator_not_readable(tmpdir):
    p = tmpdir.join("nonreadable")
    p.write("")
    os.chmod(p.strpath, stat.S_IWRITE)
    validator = FileIsReadableValidator()
    with pytest.raises(ConfigurationError) as e:
        validator(p.strpath, "path")
    assert "is not readable" in e.value.args[0]


def test_file_is_readable_validator_all_good(tmpdir):
    p = tmpdir.join("readable")
    p.write("")
    validator = FileIsReadableValidator()
    assert validator(p.strpath, "path") == p.strpath


def test_validate_catches_type_errors():
    class MyConfig(_ConfigBase):
        an_int = _ConfigValue("anint", type=int)

    c = MyConfig({"anint": "x"})
    assert "invalid literal" in c.errors["anint"]


@pytest.mark.parametrize(
    "val,expected",
    [("transactions", constants.TRANSACTION), ("errors", constants.ERROR), ("all", "all"), ("off", "off")],
)
def test_capture_body_mapping(val, expected):
    c = Config(inline_dict={"capture_body": val})
    assert c.capture_body == expected


@pytest.mark.parametrize(
    "enabled,recording,is_recording",
    [(True, True, True), (True, False, False), (False, True, False), (False, False, False)],
)
def test_is_recording(enabled, recording, is_recording):
    c = Config(inline_dict={"enabled": enabled, "recording": recording, "service_name": "foo"})
    assert c.is_recording is is_recording


def test_required_is_checked_if_field_not_provided():
    class MyConfig(_ConfigBase):
        this_one_is_required = _ConfigValue("this_one_is_required", type=int, required=True)
        this_one_isnt = _ConfigValue("this_one_isnt", type=int, required=False)

    assert MyConfig({"this_one_is_required": None}).errors
    assert MyConfig({}).errors
    assert MyConfig({"this_one_isnt": 1}).errors

    c = MyConfig({"this_one_is_required": 1})
    c.update({"this_one_isnt": 0})
    assert not c.errors


def test_callback():
    test_var = {"foo": 0}

    def set_global(dict_key, old_value, new_value, config_instance):
        # TODO make test_var `nonlocal` once we drop py2 -- it can just be a
        # basic variable then instead of a dictionary
        test_var[dict_key] += 1

    class MyConfig(_ConfigBase):
        foo = _ConfigValue("foo", callbacks=[set_global])

    c = MyConfig({"foo": "bar"})
    assert test_var["foo"] == 1
    c.update({"foo": "baz"})
    assert test_var["foo"] == 2


def test_callbacks_on_default():
    test_var = {"foo": 0}

    def set_global(dict_key, old_value, new_value, config_instance):
        # TODO make test_var `nonlocal` once we drop py2 -- it can just be a
        # basic variable then instead of a dictionary
        test_var[dict_key] += 1

    class MyConfig(_ConfigBase):
        foo = _ConfigValue("foo", callbacks=[set_global], default="foobar")

    c = MyConfig()
    assert test_var["foo"] == 1
    c = MyConfig({"foo": "bar"})
    assert test_var["foo"] == 2
    c.update({"foo": "baz"})
    assert test_var["foo"] == 3

    # Test without callback on default
    class MyConfig(_ConfigBase):
        foo = _ConfigValue("foo", callbacks=[set_global], callbacks_on_default=False, default="foobar")

    c = MyConfig()
    assert test_var["foo"] == 3
    c = MyConfig({"foo": "bar"})
    assert test_var["foo"] == 4
    c.update({"foo": "baz"})
    assert test_var["foo"] == 5


def test_callback_reset():
    test_var = {"foo": 0}

    def set_global(dict_key, old_value, new_value, config_instance):
        # TODO make test_var `nonlocal` once we drop py2 -- it can just be a
        # basic variable then instead of a dictionary
        test_var[dict_key] += 1

    class MyConfig(_ConfigBase):
        foo = _ConfigValue("foo", callbacks=[set_global])

    c = VersionedConfig(MyConfig({"foo": "bar"}), version=None)
    assert test_var["foo"] == 1
    c.update(version=2, **{"foo": "baz"})
    assert test_var["foo"] == 2
    c.reset()
    assert test_var["foo"] == 3


def test_reset_after_adding_config():
    class MyConfig(_ConfigBase):
        foo = _ConfigValue("foo")
        bar = _ConfigValue("bar")

    c = VersionedConfig(MyConfig({"foo": "baz"}), version=1)

    c.update(version=2, bar="bazzinga")

    c.reset()
    assert c.bar is None


def test_valid_values_validator():
    # Case sensitive
    v = EnumerationValidator(["foo", "Bar", "baz"], case_sensitive=False)
    assert v("foo", "foo") == "foo"
    assert v("bar", "foo") == "Bar"
    assert v("BAZ", "foo") == "baz"
    with pytest.raises(ConfigurationError):
        v("foobar", "foo")

    # Case insensitive
    v = EnumerationValidator(["foo", "Bar", "baz"], case_sensitive=True)
    assert v("foo", "foo") == "foo"
    with pytest.raises(ConfigurationError):
        v("bar", "foo")
    with pytest.raises(ConfigurationError):
        v("BAZ", "foo")
    assert v("Bar", "foo") == "Bar"
    with pytest.raises(ConfigurationError):
        v("foobar", "foo")


def test_versioned_config_attribute_access(elasticapm_client):
    # see https://github.com/elastic/apm-agent-python/issues/1147
    val = elasticapm_client.config.start_stop_order
    assert isinstance(val, int)
    # update config to ensure start_stop_order isn't read from the proxied Config object
    elasticapm_client.config.update("2", capture_body=True)
    val = elasticapm_client.config.start_stop_order
    assert isinstance(val, int)


def test_config_all_upper_case():
    c = Config.__class__.__dict__.items()
    for field, config_value in Config.__dict__.items():
        if not isinstance(config_value, _ConfigValue):
            continue
        assert config_value.env_key == config_value.env_key.upper()


def test_regex_validator_without_match():
    validator = RegexValidator(r"\d")
    with pytest.raises(ConfigurationError) as e:
        validator("foo", "field")
    assert "does not match pattern" in e.value.args[0]


def test_unit_validator_without_match():
    validator = RegexValidator("ms")
    with pytest.raises(ConfigurationError) as e:
        validator("s", "field")
    assert "does not match pattern" in e.value.args[0]


def test_unit_validator_with_unsupported_unit():
    validator = UnitValidator(r"(\d+)(s)", "secs", {})
    with pytest.raises(ConfigurationError) as e:
        validator("10s", "field")
    assert "is not a supported unit" in e.value.args[0]


def test_precision_validator_not_a_float():
    validator = PrecisionValidator()
    with pytest.raises(ConfigurationError) as e:
        validator("notafloat", "field")
    assert "is not a float" in e.value.args[0]


def test_exclude_range_validator_not_in_range():
    validator = ExcludeRangeValidator(1, 100, "desc")
    with pytest.raises(ConfigurationError) as e:
        validator(10, "field")
    assert "cannot be in range" in e.value.args[0]


def test_supported_value_in_fips_mode_validator_in_fips_mode_with_invalid_value(monkeypatch):
    monkeypatch.setattr(elasticapm.conf, "_in_fips_mode", lambda: True)
    exception_message = "VERIFY_SERVER_CERT=False must be set to True if FIPS mode is enabled"
    validator = SupportedValueInFipsModeValidator(supported_value=True)
    with pytest.raises(ConfigurationError) as e:
        validator(False, "VERIFY_SERVER_CERT")
    assert exception_message == e.value.args[0]

    config = Config({"VERIFY_SERVER_CERT": False})
    assert config.errors["VERIFY_SERVER_CERT"] == exception_message


def test_supported_value_in_fips_mode_validator_in_fips_mode_with_valid_value(monkeypatch):
    monkeypatch.setattr(elasticapm.conf, "_in_fips_mode", lambda: True)
    validator = SupportedValueInFipsModeValidator(supported_value=True)
    assert validator(True, "VERIFY_SERVER_CERT") == True
    config = Config({"VERIFY_SERVER_CERT": True})
    assert config.verify_server_cert == True
    assert "VERIFY_SERVER_CERT" not in config.errors


def test_supported_value_in_fips_mode_validator_not_in_fips_mode(monkeypatch):
    monkeypatch.setattr(elasticapm.conf, "_in_fips_mode", lambda: False)
    validator = SupportedValueInFipsModeValidator(supported_value=True)
    assert validator(True, "field") == True
    assert validator(False, "field") == False

    config = Config({"VERIFY_SERVER_CERT": False})
    assert not config.errors
    config = Config({"VERIFY_SERVER_CERT": True})
    assert not config.errors



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import timedelta

import pytest  # isort:skip

opentracing = pytest.importorskip("opentracing")  # isort:skip

import sys

import mock
from opentracing import Format

import elasticapm
from elasticapm.conf import constants
from elasticapm.contrib.opentracing import Tracer
from elasticapm.contrib.opentracing.span import OTSpanContext
from elasticapm.utils.disttracing import TraceParent

pytestmark = pytest.mark.opentracing


try:
    from opentracing import logs as ot_logs
    from opentracing import tags
except ImportError:
    ot_logs = None


@pytest.fixture()
def tracer(elasticapm_client):
    yield Tracer(client_instance=elasticapm_client)
    elasticapm.uninstrument()


def test_tracer_with_instantiated_client(elasticapm_client):
    tracer = Tracer(client_instance=elasticapm_client)
    assert tracer._agent is elasticapm_client


def test_tracer_with_config():
    config = {"METRICS_INTERVAL": "0s", "SERVER_URL": "https://example.com/test"}
    tracer = Tracer(config=config)
    try:
        assert tracer._agent.config.metrics_interval == timedelta(seconds=0)
        assert tracer._agent.config.server_url == "https://example.com/test"
    finally:
        tracer._agent.close()


def test_tracer_instrument(elasticapm_client):
    with mock.patch("elasticapm.contrib.opentracing.tracer.instrument") as mock_instrument:
        elasticapm_client.config.instrument = False
        Tracer(client_instance=elasticapm_client)
        assert mock_instrument.call_count == 0

        elasticapm_client.config.instrument = True
        Tracer(client_instance=elasticapm_client)
        assert mock_instrument.call_count == 1


def test_ot_transaction_started(tracer):
    with tracer.start_active_span("test") as ot_scope:
        ot_scope.span.set_tag("result", "OK")
    client = tracer._agent
    transaction = client.events[constants.TRANSACTION][0]
    assert transaction["type"] == "custom"
    assert transaction["name"] == "test"
    assert transaction["result"] == "OK"


def test_ot_span(tracer):
    with tracer.start_active_span("test") as ot_scope_transaction:
        with tracer.start_active_span("testspan") as ot_scope_span:
            ot_scope_span.span.set_tag("span.kind", "custom")
            with tracer.start_active_span("testspan2") as ot_scope_span2:
                with tracer.start_active_span("testspan3", child_of=ot_scope_span.span) as ot_scope_span3:
                    pass
    client = tracer._agent
    transaction = client.events[constants.TRANSACTION][0]
    span1 = client.events[constants.SPAN][2]
    span2 = client.events[constants.SPAN][1]
    span3 = client.events[constants.SPAN][0]
    assert span1["transaction_id"] == span1["parent_id"] == transaction["id"]
    assert span1["name"] == "testspan"

    assert span2["transaction_id"] == transaction["id"]
    assert span2["parent_id"] == span1["id"]
    assert span2["name"] == "testspan2"

    # check that span3 has span1 as parent
    assert span3["transaction_id"] == transaction["id"]
    assert span3["parent_id"] == span1["id"]
    assert span3["name"] == "testspan3"


def test_transaction_tags(tracer):
    with tracer.start_active_span("test") as ot_scope:
        ot_scope.span.set_tag("type", "foo")
        ot_scope.span.set_tag("http.status_code", 200)
        ot_scope.span.set_tag("http.url", "http://example.com/foo")
        ot_scope.span.set_tag("http.method", "GET")
        ot_scope.span.set_tag("user.id", 1)
        ot_scope.span.set_tag("user.email", "foo@example.com")
        ot_scope.span.set_tag("user.username", "foo")
        ot_scope.span.set_tag("component", "Django")
        ot_scope.span.set_tag("something.else", "foo")
    client = tracer._agent
    transaction = client.events[constants.TRANSACTION][0]

    assert transaction["type"] == "foo"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["context"]["response"]["status_code"] == 200
    assert transaction["context"]["request"]["url"]["full"] == "http://example.com/foo"
    assert transaction["context"]["request"]["method"] == "GET"
    assert transaction["context"]["user"] == {"id": 1, "email": "foo@example.com", "username": "foo"}
    assert transaction["context"]["service"]["framework"]["name"] == "Django"
    assert transaction["context"]["tags"] == {"something_else": "foo"}


def test_span_tags(tracer):
    with tracer.start_active_span("transaction") as ot_scope_t:
        with tracer.start_active_span("span") as ot_scope_s:
            s = ot_scope_s.span
            s.set_tag("db.type", "sql")
            s.set_tag("db.statement", "SELECT * FROM foo")
            s.set_tag("db.user", "bar")
            s.set_tag("db.instance", "baz")
        with tracer.start_active_span("span") as ot_scope_s:
            s = ot_scope_s.span
            s.set_tag("span.kind", "foo")
            s.set_tag("something.else", "bar")
    client = tracer._agent
    span1 = client.events[constants.SPAN][0]
    span2 = client.events[constants.SPAN][1]

    assert span1["context"]["db"] == {"type": "sql", "user": "bar", "statement": "SELECT * FROM foo"}
    assert span1["type"] == "db.sql"
    assert span1["context"]["tags"] == {"db_instance": "baz"}

    assert span2["type"] == "foo"
    assert span2["context"]["tags"] == {"something_else": "bar"}


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 1}], indirect=True)
def test_dropped_spans(tracer):
    assert tracer._agent.config.transaction_max_spans == 1
    with tracer.start_active_span("transaction") as ot_scope_t:
        with tracer.start_active_span("span") as ot_scope_s:
            s = ot_scope_s.span
            s.set_tag("db.type", "sql")
        with tracer.start_active_span("span") as ot_scope_s:
            s = ot_scope_s.span
            s.set_tag("db.type", "sql")
    client = tracer._agent
    spans = client.events[constants.SPAN]
    assert len(spans) == 1


def test_error_log(tracer):
    with tracer.start_active_span("transaction") as tx_scope:
        try:
            raise ValueError("oops")
        except ValueError:
            exc_type, exc_val, exc_tb = sys.exc_info()[:3]
            tx_scope.span.log_kv(
                {"python.exception.type": exc_type, "python.exception.val": exc_val, "python.exception.tb": exc_tb}
            )
    client = tracer._agent
    error = client.events[constants.ERROR][0]

    assert error["exception"]["message"] == "ValueError: oops"


@pytest.mark.skipif(ot_logs is None, reason="New key names in opentracing-python 2.1")
def test_error_log_ot_21(tracer):
    with tracer.start_active_span("transaction") as tx_scope:
        try:
            raise ValueError("oops")
        except ValueError:
            exc_type, exc_val, exc_tb = sys.exc_info()[:3]
            tx_scope.span.log_kv(
                {
                    ot_logs.EVENT: tags.ERROR,
                    ot_logs.ERROR_KIND: exc_type,
                    ot_logs.ERROR_OBJECT: exc_val,
                    ot_logs.STACK: exc_tb,
                }
            )
    client = tracer._agent
    error = client.events[constants.ERROR][0]

    assert error["exception"]["message"] == "ValueError: oops"


def test_error_log_automatic_in_span_context_manager(tracer):
    scope = tracer.start_active_span("transaction")
    with pytest.raises(ValueError):
        with scope.span:
            raise ValueError("oops")

    client = tracer._agent
    error = client.events[constants.ERROR][0]

    assert error["exception"]["message"] == "ValueError: oops"


def test_span_set_bagge_item_noop(tracer):
    scope = tracer.start_active_span("transaction")
    assert scope.span.set_baggage_item("key", "val") == scope.span


def test_tracer_extract_http(tracer):
    span_context = tracer.extract(
        Format.HTTP_HEADERS, {"elastic-apm-traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"}
    )

    assert span_context.trace_parent.version == 0
    assert span_context.trace_parent.trace_id == "0af7651916cd43dd8448eb211c80319c"
    assert span_context.trace_parent.span_id == "b7ad6b7169203331"


def test_tracer_extract_map(tracer):
    span_context = tracer.extract(
        Format.TEXT_MAP, {"elastic-apm-traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"}
    )

    assert span_context.trace_parent.version == 0
    assert span_context.trace_parent.trace_id == "0af7651916cd43dd8448eb211c80319c"
    assert span_context.trace_parent.span_id == "b7ad6b7169203331"


def test_tracer_extract_binary(tracer):
    with pytest.raises(opentracing.UnsupportedFormatException):
        tracer.extract(Format.BINARY, b"foo")


def test_tracer_extract_corrupted(tracer):
    with pytest.raises(opentracing.SpanContextCorruptedException):
        tracer.extract(Format.HTTP_HEADERS, {"nothing-to": "see-here"})


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
def test_tracer_inject_http(tracer):
    span_context = OTSpanContext(
        trace_parent=TraceParent.from_string("00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01")
    )
    carrier = {}
    tracer.inject(span_context, Format.HTTP_HEADERS, carrier)
    assert carrier[constants.TRACEPARENT_HEADER_NAME] == b"00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"
    if tracer._agent.config.use_elastic_traceparent_header:
        assert carrier[constants.TRACEPARENT_LEGACY_HEADER_NAME] == carrier[constants.TRACEPARENT_HEADER_NAME]


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
def test_tracer_inject_map(tracer):
    span_context = OTSpanContext(
        trace_parent=TraceParent.from_string("00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01")
    )
    carrier = {}
    tracer.inject(span_context, Format.TEXT_MAP, carrier)
    assert carrier[constants.TRACEPARENT_HEADER_NAME] == b"00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"
    if tracer._agent.config.use_elastic_traceparent_header:
        assert carrier[constants.TRACEPARENT_LEGACY_HEADER_NAME] == carrier[constants.TRACEPARENT_HEADER_NAME]


def test_tracer_inject_binary(tracer):
    span_context = OTSpanContext(
        trace_parent=TraceParent.from_string("00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01")
    )
    with pytest.raises(opentracing.UnsupportedFormatException):
        tracer.inject(span_context, Format.BINARY, {})



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from asyncio import tasks

import pytest

import elasticapm
from elasticapm.conf import constants

pytestmark = [pytest.mark.asyncio]


async def test_async_capture_span(instrument, elasticapm_client):
    @elasticapm.async_capture_span()
    async def do_some_work():
        async with elasticapm.async_capture_span("more-work"):
            await tasks.sleep(0.1)

    elasticapm_client.begin_transaction("test")
    await do_some_work()
    elasticapm_client.end_transaction("test", "OK")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 3
    sleep, c, d = spans
    assert sleep["subtype"] == "sleep"
    assert not sleep["sync"]
    assert sleep["transaction_id"] == transaction["id"]

    assert c["type"] == "code"
    assert c["subtype"] == "custom"
    assert not c["sync"]
    assert c["transaction_id"] == transaction["id"]

    assert d["type"] == "code"
    assert d["subtype"] == "custom"
    assert not d["sync"]
    assert d["transaction_id"] == transaction["id"]



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE

from string import ascii_uppercase
from typing import Dict

from sanic import Sanic
from sanic import __version__ as version
from sanic.cookies import Cookie, CookieJar
from sanic.request import Request
from sanic.response import HTTPResponse

from elasticapm.base import Client
from elasticapm.conf import Config, constants
from elasticapm.contrib.sanic.sanic_types import EnvInfoType
from elasticapm.utils import get_url_dict


class SanicAPMConfig(dict):
    def __init__(self, app: Sanic) -> None:
        super(SanicAPMConfig, self).__init__()
        for _key, _v in app.config.items():
            if _key.startswith("ELASTIC_APM_"):
                self[_key.replace("ELASTIC_APM_", "")] = _v


def get_env(request: Request) -> EnvInfoType:
    """
    Extract Server Environment Information from the current Request's context
    :param request: Inbound HTTP Request
    :return: A tuple containing the attribute and it's corresponding value for the current Application ENV
    """
    for _attr in ("server_name", "server_port", "version"):
        if hasattr(request, _attr):
            yield _attr, getattr(request, _attr)


# noinspection PyBroadException
async def get_request_info(config: Config, request: Request, event_type: str) -> Dict[str, str]:
    """
    Generate a traceable context information from the inbound HTTP request

    :param config: Application Configuration used to tune the way the data is captured
    :param request: Inbound HTTP request
    :param event_type: the event type (such as constants.TRANSACTION) for determing whether to capture the body
    :return: A dictionary containing the context information of the ongoing transaction
    """
    env = dict(get_env(request=request))
    app_config = {k: v for k, v in dict(request.app.config).items() if all(letter in ascii_uppercase for letter in k)}
    env.update(app_config)
    result = {
        "env": env,
        "method": request.method,
        "socket": {
            "remote_address": _get_client_ip(request=request),
            "encrypted": request.scheme in ["https", "wss"],
        },
        "cookies": request.cookies,
        "http_version": request.version,
    }
    if config.capture_headers:
        result["headers"] = dict(request.headers)

    if request.method in constants.HTTP_WITH_BODY and config.capture_body in ("all", event_type):
        if request.content_type.startswith("multipart") or "octet-stream" in request.content_type:
            result["body"] = "[DISCARDED]"
        try:
            result["body"] = request.body.decode("utf-8")
        except Exception:
            pass

    if "body" not in result:
        result["body"] = "[REDACTED]"
    result["url"] = get_url_dict(request.url)
    return result


async def get_response_info(config: Config, response: HTTPResponse, event_type: str) -> Dict[str, str]:
    """
    Generate a traceable context information from the inbound HTTP Response

    :param config: Application Configuration used to tune the way the data is captured
    :param response: outbound HTTP Response
    :param event_type: the event type (such as constants.TRANSACTION) for determing whether to capture the body
    :return: A dictionary containing the context information of the ongoing transaction
    """
    result = {
        "cookies": _transform_response_cookie(cookies=response.cookies),
        "finished": True,
        "headers_sent": True,
    }
    if isinstance(response.status, int):
        result["status_code"] = response.status

    if config.capture_headers:

        def normalize(v):
            # we are getting entries for Set-Cookie headers as Cookie instances
            if isinstance(v, Cookie):
                return str(v)
            return v

        result["headers"] = {k: normalize(v) for k, v in response.headers.items()}

    if config.capture_body in ("all", event_type) and "octet-stream" not in response.content_type:
        result["body"] = response.body.decode("utf-8")
    else:
        result["body"] = "[REDACTED]"

    return result


def _get_client_ip(request: Request) -> str:
    """Extract Client IP Address Information"""
    try:
        return request.ip or request.socket[0] or request.remote_addr
    except IndexError:
        return request.remote_addr


def make_client(client_cls=Client, **defaults) -> Client:
    if "framework_name" not in defaults:
        defaults["framework_name"] = "sanic"
        defaults["framework_version"] = version

    return client_cls(**defaults)


def _transform_response_cookie(cookies: CookieJar) -> Dict[str, str]:
    """Transform the Sanic's CookieJar instance into a Normal dictionary to build the context"""
    # old sanic versions used to have an items() method
    if hasattr(cookies, "items"):
        return {k: {"value": v.value, "path": v["path"]} for k, v in cookies.items()}

    try:
        return {cookie.key: {"value": cookie.value, "path": cookie.path} for cookie in cookies.cookies}
    except KeyError:
        # cookies.cookies assumes Set-Cookie header will be there
        return {}



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("cassandra")  # isort:skip

import os
import socket

from cassandra.cluster import Cluster
from cassandra.protocol import ConfigurationException
from cassandra.query import SimpleStatement

from elasticapm.conf.constants import TRANSACTION
from elasticapm.instrumentation.packages.dbapi2 import extract_signature

pytestmark = [pytest.mark.cassandra]

if "CASSANDRA_HOST" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping cassandra tests, no CASSANDRA_HOST environment variable set"))


@pytest.fixture()
def cassandra_cluster():
    cluster = Cluster([os.environ.get("CASSANDRA_HOST", "localhost")])
    yield cluster
    del cluster


@pytest.fixture()
def cassandra_session(cassandra_cluster):
    session = cassandra_cluster.connect()
    try:
        # make sure the keyspace doesn't exist
        session.execute("DROP KEYSPACE testkeyspace;")
    except ConfigurationException:
        pass
    session.execute("""
        CREATE KEYSPACE testkeyspace
        WITH REPLICATION = { 'class' : 'SimpleStrategy' ,'replication_factor' : 1 }
    """)
    session.execute("USE testkeyspace;")
    session.execute("CREATE TABLE testkeyspace.users ( id UUID PRIMARY KEY, name text);")
    yield session
    session.execute("DROP KEYSPACE testkeyspace;")


def test_cassandra_connect(instrument, elasticapm_client, cassandra_cluster):
    elasticapm_client.begin_transaction("transaction.test")
    sess = cassandra_cluster.connect()
    elasticapm_client.end_transaction("test")

    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert span["type"] == "db"
    assert span["subtype"] == "cassandra"
    assert span["action"] == "connect"
    assert span["duration"] > 0
    assert span["name"] == "Cluster.connect"
    assert span["context"]["destination"] == {
        "address": socket.gethostbyname(os.environ.get("CASSANDRA_HOST", "localhost")),
        "port": 9042,
        "service": {"name": "", "resource": "cassandra", "type": ""},
    }


def test_cassandra_connect_keyspace(instrument, elasticapm_client, cassandra_cluster):
    session = cassandra_cluster.connect()
    try:
        session.execute("""
            CREATE KEYSPACE testkeyspace
            WITH REPLICATION = { 'class' : 'SimpleStrategy' ,'replication_factor' : 1 }
        """)
        elasticapm_client.begin_transaction("transaction.test")
        sess = cassandra_cluster.connect("testkeyspace")
        elasticapm_client.end_transaction("test")
    finally:
        session.execute("DROP KEYSPACE testkeyspace;")

    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert span["type"] == "db"
    assert span["subtype"] == "cassandra"
    assert span["action"] == "connect"
    assert span["duration"] > 0
    assert span["name"] == "Cluster.connect"
    assert span["context"]["destination"] == {
        "address": socket.gethostbyname(os.environ.get("CASSANDRA_HOST", "localhost")),
        "port": 9042,
        "service": {"name": "", "resource": "cassandra/testkeyspace", "type": ""},
    }
    assert span["context"]["service"]["target"]["type"] == "cassandra"
    assert span["context"]["service"]["target"]["name"] == "testkeyspace"


def test_select_query_string(instrument, cassandra_session, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    cassandra_session.execute("SELECT name from users")
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["type"] == "db"
    assert span["subtype"] == "cassandra"
    assert span["action"] == "query"
    assert span["name"] == "SELECT FROM users"
    assert span["context"]["db"] == {"statement": "SELECT name from users", "type": "sql", "instance": "testkeyspace"}
    assert span["context"]["destination"] == {
        "address": socket.gethostbyname(os.environ.get("CASSANDRA_HOST", "localhost")),
        "port": 9042,
        "service": {"name": "", "resource": "cassandra/testkeyspace", "type": ""},
    }
    assert span["context"]["service"]["target"]["type"] == "cassandra"
    assert span["context"]["service"]["target"]["name"] == "testkeyspace"


def test_select_simple_statement(instrument, cassandra_session, elasticapm_client):
    statement = SimpleStatement("SELECT name from users")
    elasticapm_client.begin_transaction("transaction.test")
    cassandra_session.execute(statement)
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["type"] == "db"
    assert span["subtype"] == "cassandra"
    assert span["action"] == "query"
    assert span["name"] == "SELECT FROM users"
    assert span["context"]["db"] == {"statement": "SELECT name from users", "type": "sql", "instance": "testkeyspace"}


def test_select_prepared_statement(instrument, cassandra_session, elasticapm_client):
    prepared_statement = cassandra_session.prepare("SELECT name from users")
    elasticapm_client.begin_transaction("transaction.test")
    cassandra_session.execute(prepared_statement)
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["type"] == "db"
    assert span["subtype"] == "cassandra"
    assert span["action"] == "query"
    assert span["name"] == "SELECT FROM users"
    assert span["context"]["db"] == {"statement": "SELECT name from users", "type": "sql", "instance": "testkeyspace"}


def test_signature_create_keyspace():
    assert (
        extract_signature(
            "CREATE KEYSPACE testkeyspace WITH REPLICATION = { 'class' : 'NetworkTopologyStrategy', 'datacenter1' : 3 };"
        )
        == "CREATE KEYSPACE"
    )


def test_signature_create_columnfamily():
    assert extract_signature("""CREATE COLUMNFAMILY users (
  userid text PRIMARY KEY,
  first_name text,
  last_name text,
  emails set<text>,
  top_scores list<int>,
  todo map<timestamp, text>
);""") == "CREATE COLUMNFAMILY"


def test_select_from_collection():
    assert extract_signature("SELECT first, last FROM a.b WHERE id = 1;") == "SELECT FROM a.b"


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "50ms",
        }
    ],
    indirect=True,
)
def test_cassandra_span_compression(instrument, cassandra_session, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    for i in range(5):
        cassandra_session.execute("SELECT name from users")
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from typing import cast

import pytest

from elasticapm import get_client
from elasticapm.conf.constants import SPAN, TRANSACTION
from elasticapm.instrumentation.packages.psycopg import PGCursorProxy
from elasticapm.utils import default_ports
from tests.fixtures import TempStoreClient

psycopg = pytest.importorskip("psycopg")

pytestmark = pytest.mark.psycopg

has_postgres_configured = "POSTGRES_DB" in os.environ

PSYCOPG_VERSION = tuple([int(x) for x in psycopg.version.__version__.split() if x.isdigit()])


def connect_kwargs():
    return {
        "dbname": os.environ.get("POSTGRES_DB", "elasticapm_test"),
        "user": os.environ.get("POSTGRES_USER", "postgres"),
        "password": os.environ.get("POSTGRES_PASSWORD", "postgres"),
        "host": os.environ.get("POSTGRES_HOST", None),
        "port": os.environ.get("POSTGRES_PORT", None),
    }


@pytest.fixture(scope="function")
def postgres_connection(request):
    conn = psycopg.connect(**connect_kwargs())
    cursor = conn.cursor()
    cursor.execute(
        "CREATE TABLE test(id int, name VARCHAR(5) NOT NULL);"
        "INSERT INTO test VALUES (1, 'one'), (2, 'two'), (3, 'three');"
    )

    yield conn

    # cleanup
    cursor.execute("ROLLBACK")


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_cursor_execute_signature(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    cursor.execute(query="SELECT 1", params=None, prepare=None, binary=None)
    row = cursor.fetchone()
    assert row


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_cursor_executemany_signature(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    if PSYCOPG_VERSION < (3, 1, 0):
        kwargs = {}
    else:
        kwargs = {"returning": False}
    res = cursor.executemany(
        query="INSERT INTO test VALUES (%s, %s)",
        params_seq=((4, "four"),),
        **kwargs,
    )
    assert res is None


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_destination(instrument, postgres_connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    cursor = postgres_connection.cursor()
    cursor.execute("SELECT 1")
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["context"]["destination"] == {
        "address": os.environ.get("POSTGRES_HOST", None),
        "port": default_ports["postgresql"],
        "service": {"name": "", "resource": "postgresql/elasticapm_test", "type": ""},
    }


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_tracing_outside_of_elasticapm_transaction(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    # check that the cursor is a proxy, even though we're not in an elasticapm
    # transaction
    assert isinstance(cursor, PGCursorProxy)
    cursor.execute("SELECT 1")
    transactions = elasticapm_client.events[TRANSACTION]
    assert not transactions


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_select_LIKE(instrument, postgres_connection, elasticapm_client):
    """
    Check that we pass queries with %-notation but without parameters
    properly to the dbapi backend
    """
    cursor = postgres_connection.cursor()
    query = "SELECT * FROM test WHERE name LIKE 't%'"

    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        cursor.fetchall()
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the spans for the other tests.
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert span["type"] == "db"
        assert span["subtype"] == "postgresql"
        assert span["action"] == "query"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "elasticapm_test"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == query
        assert span["context"]["service"]["target"]["type"] == "postgresql"
        assert span["context"]["service"]["target"]["name"] == "elasticapm_test"


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_composable_query_works(instrument, postgres_connection, elasticapm_client):
    """
    Check that we parse queries that are psycopg.sql.Composable correctly
    """
    from psycopg import sql

    cursor = postgres_connection.cursor()
    query = sql.SQL("SELECT * FROM {table} WHERE {row} LIKE 't%' ORDER BY {row} DESC").format(
        table=sql.Identifier("test"), row=sql.Identifier("name")
    )
    baked_query = query.as_string(cursor.__wrapped__)
    result = None
    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        result = cursor.fetchall()
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the spans for the other tests.
        assert [(2, "two"), (3, "three")] == result
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "elasticapm_test"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == baked_query


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_binary_query_works(instrument, postgres_connection, elasticapm_client):
    """
    Check that we pass queries with %-notation but without parameters
    properly to the dbapi backend
    """
    cursor = postgres_connection.cursor()
    query = b"SELECT * FROM test WHERE name LIKE 't%'"

    baked_query = query.decode()
    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute(query)
        result = cursor.fetchall()
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the spans for the other tests.
        assert [(2, "two"), (3, "three")] == result
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])
        span = spans[0]
        assert span["name"] == "SELECT FROM test"
        assert "db" in span["context"]
        assert span["context"]["db"]["instance"] == "elasticapm_test"
        assert span["context"]["db"]["type"] == "sql"
        assert span["context"]["db"]["statement"] == baked_query


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_context_manager(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    with psycopg.connect(**connect_kwargs()) as conn:
        with conn.cursor() as curs:
            curs.execute("SELECT 1;")
            curs.fetchall()
    elasticapm_client.end_transaction("test", "OK")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert len(spans) == 2
    assert spans[0]["subtype"] == "postgresql"
    assert spans[0]["action"] == "connect"
    assert spans[0]["context"]["service"]["target"]["type"] == "postgresql"
    assert spans[0]["context"]["service"]["target"]["name"] == "elasticapm_test"

    assert spans[1]["subtype"] == "postgresql"
    assert spans[1]["action"] == "query"


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_psycopg_rows_affected(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor()
    try:
        elasticapm_client.begin_transaction("web.django")
        cursor.execute("INSERT INTO test VALUES (4, 'four')")
        cursor.execute("SELECT * FROM test")
        cursor.execute("UPDATE test SET name = 'five' WHERE  id = 4")
        cursor.execute("DELETE FROM test WHERE  id = 4")
        elasticapm_client.end_transaction(None, "test-transaction")
    finally:
        transactions = elasticapm_client.events[TRANSACTION]
        spans = elasticapm_client.spans_for_transaction(transactions[0])

        assert spans[0]["name"] == "INSERT INTO test"
        assert spans[0]["context"]["db"]["rows_affected"] == 1

        assert spans[1]["name"] == "SELECT FROM test"
        assert "rows_affected" not in spans[1]["context"]["db"]

        assert spans[2]["name"] == "UPDATE test"
        assert spans[2]["context"]["db"]["rows_affected"] == 1

        assert spans[3]["name"] == "DELETE FROM test"
        assert spans[3]["context"]["db"]["rows_affected"] == 1


@pytest.mark.integrationtest
def test_psycopg_connection(instrument, elasticapm_transaction, postgres_connection):
    # elastciapm_client.events is only available on `TempStoreClient`, this keeps the type checkers happy
    elasticapm_client = cast(TempStoreClient, get_client())
    elasticapm_client.end_transaction("test", "success")
    span = elasticapm_client.events[SPAN][0]
    host = os.environ.get("POSTGRES_HOST", "localhost")
    assert span["name"] == f"psycopg.connect {host}:5432"
    assert span["action"] == "connect"


@pytest.mark.integrationtest
@pytest.mark.skipif(not has_postgres_configured, reason="PostgresSQL not configured")
def test_server_cursor_execute(instrument, postgres_connection, elasticapm_client):
    cursor = postgres_connection.cursor(name="server_cursor")
    assert isinstance(cursor, psycopg.ServerCursor)
    record = cursor.execute(query="SELECT 1", params=None, binary=True)
    assert record



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest

azure = pytest.importorskip("azure.functions")

import datetime
import os

import azure.functions as func
import mock

import elasticapm
from elasticapm.conf import constants
from elasticapm.contrib.serverless.azure import AzureFunctionsClient, ElasticAPMExtension, get_faas_data
from tests.fixtures import TempStoreClient


class AzureFunctionsTestClient(AzureFunctionsClient, TempStoreClient):
    pass


@pytest.mark.parametrize(
    "elasticapm_client", [{"client_class": AzureFunctionsTestClient}], indirect=["elasticapm_client"]
)
def test_service_info(elasticapm_client):
    with mock.patch.dict(
        os.environ,
        {
            "FUNCTIONS_EXTENSION_VERSION": "1.1",
            "FUNCTIONS_WORKER_RUNTIME": "MontyPython",
            "FUNCTIONS_WORKER_RUNTIME_VERSION": "2.2",
            "WEBSITE_INSTANCE_ID": "foo",
        },
    ):
        service_info = elasticapm_client.get_service_info()
        assert service_info["framework"]["name"] == "Azure Functions"
        assert service_info["framework"]["version"] == "1.1"
        assert service_info["runtime"]["name"] == "MontyPython"
        assert service_info["runtime"]["version"] == "2.2"
        assert service_info["node"]["configured_name"] == "foo"


@pytest.mark.parametrize(
    "elasticapm_client", [{"client_class": AzureFunctionsTestClient}], indirect=["elasticapm_client"]
)
def test_cloud_info(elasticapm_client):
    with mock.patch.dict(
        os.environ,
        {
            "REGION_NAME": "eu-liechtenstein",
            "WEBSITE_OWNER_NAME": "2491fc8e-f7c1-4020-b9c6-78509919fd16+my-resource-group-ARegionShortNamewebspace",
            "WEBSITE_SITE_NAME": "foo",
            "WEBSITE_RESOURCE_GROUP": "bar",
        },
    ):
        cloud_info = elasticapm_client.get_cloud_info()
        assert cloud_info["provider"] == "azure"
        assert cloud_info["region"] == "eu-liechtenstein"
        assert cloud_info["service"]["name"] == "functions"
        assert cloud_info["account"]["id"] == "2491fc8e-f7c1-4020-b9c6-78509919fd16"
        assert cloud_info["instance"]["name"] == "foo"
        assert cloud_info["project"]["name"] == "bar"


def test_extension_configure():
    try:
        with mock.patch.dict(os.environ, {"WEBSITE_SITE_NAME": "foo", "AZURE_FUNCTIONS_ENVIRONMENT": "prod"}):
            ElasticAPMExtension.configure(client_class=AzureFunctionsTestClient)
        client = ElasticAPMExtension.client
        assert client.config.metrics_interval == datetime.timedelta(0)
        assert client.config.breakdown_metrics is False
        assert client.config.central_config is False
        assert client.config.cloud_provider == "none"
        assert client.config.framework_name == "Azure Functions"
        assert client.config.service_name == "foo"
        assert client.config.environment == "prod"
    finally:
        if ElasticAPMExtension.client:
            ElasticAPMExtension.client.close()
            ElasticAPMExtension.client = None


def test_extension_configure_with_kwargs():
    try:
        ElasticAPMExtension.configure(
            client_class=AzureFunctionsTestClient, metrics_sets=["foo"], service_name="foo", environment="bar"
        )
        client = ElasticAPMExtension.client

        assert client.config.metrics_interval == datetime.timedelta(0)
        assert client.config.breakdown_metrics is False
        assert client.config.central_config is False
        assert client.config.cloud_provider == "none"
        assert client.config.framework_name == "Azure Functions"
        assert client.config.service_name == "foo"
        assert client.config.environment == "bar"
        assert client.config.metrics_sets == ["foo"]
    finally:
        if ElasticAPMExtension.client:
            ElasticAPMExtension.client.close()
            ElasticAPMExtension.client = None


@pytest.mark.parametrize(
    "elasticapm_client", [{"client_class": AzureFunctionsTestClient}], indirect=["elasticapm_client"]
)
def test_pre_post_invocation_app_level_request(elasticapm_client):
    try:
        ElasticAPMExtension.client = elasticapm_client
        request = func.HttpRequest(
            "get",
            "/foo/bar",
            headers={
                "Cookie": "foo=bar; baz=bazzinga",
                "traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01",
            },
            body=b"",
        )
        response = func.HttpResponse("", status_code=200, headers={}, mimetype="text/html")
        context = mock.Mock(function_name="foo_function", invocation_id="fooid")
        ElasticAPMExtension.pre_invocation_app_level(None, context, {"request": request})
        ElasticAPMExtension.post_invocation_app_level(None, context, func_ret=response)
        transaction = elasticapm_client.events[constants.TRANSACTION][0]
        assert transaction["name"] == "foo_function"
        assert transaction["type"] == "request"
        assert transaction["outcome"] == "success"
        assert transaction["trace_id"] == "0af7651916cd43dd8448eb211c80319c"
        assert transaction["parent_id"] == "b7ad6b7169203331"
        assert transaction["context"]["request"]["method"] == "GET"
        assert transaction["context"]["request"]["cookies"] == {
            "foo": "bar",
            "baz": "bazzinga",
        }
        assert transaction["context"]["request"]["url"]["full"] == "/foo/bar"
        assert transaction["context"]["response"]["status_code"] == 200

    finally:
        ElasticAPMExtension.client = None


def test_get_faas_data():
    context = mock.Mock(invocation_id="fooid", function_name="fname")
    with mock.patch.dict(
        os.environ,
        {
            "WEBSITE_OWNER_NAME": "2491fc8e-f7c1-4020-b9c6-78509919fd16+my-resource-group-ARegionShortNamewebspace",
            "WEBSITE_SITE_NAME": "foo",
            "WEBSITE_RESOURCE_GROUP": "bar",
        },
    ):
        data = get_faas_data(context, True, "request")
    assert data["coldstart"] is True
    assert data["execution"] == "fooid"
    assert data["trigger"]["type"] == "request"
    assert (
        data["id"]
        == "/subscriptions/2491fc8e-f7c1-4020-b9c6-78509919fd16/resourceGroups/bar/providers/Microsoft.Web/sites/foo/functions/fname"
    )



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

sanic = pytest.importorskip("sanic")  # isort:skip

import logging
import time
import typing as t

import pytest
from sanic import Sanic
from sanic.blueprints import Blueprint
from sanic.handlers import ErrorHandler
from sanic.request import Request
from sanic.response import HTTPResponse, json

import elasticapm
from elasticapm import async_capture_span
from elasticapm.contrib.sanic import ElasticAPM


class CustomException(Exception):
    pass


class CustomErrorHandler(ErrorHandler):
    def __init__(self) -> None:
        super(CustomErrorHandler, self).__init__()

    def default(self, request, exception):
        return json({"source": "custom-handler-default"}, 500)


@pytest.fixture()
def custom_error_handler():
    return CustomErrorHandler()


@pytest.fixture()
def sanic_elastic_app(elasticapm_client):
    def _generate(
        error_handler=None,
        elastic_client=None,
        elastic_client_cls=None,
        config=None,
        transaction_name_callback=None,
        user_context_callback=None,
        custom_context_callback=None,
        label_info_callback=None,
    ):
        Sanic.test_mode = True
        args = {"name": "elastic-apm-test-app"}
        if error_handler:
            args["error_handler"] = error_handler

        app = Sanic(**args)
        apm_args = {}
        for key, value in {
            "client": elastic_client,
            "client_cls": elastic_client_cls,
            "config": config,
            "transaction_name_callback": transaction_name_callback,
            "user_context_callback": user_context_callback,
            "custom_context_callback": custom_context_callback,
            "label_info_callback": label_info_callback,
        }.items():
            if value is not None:
                apm_args[key] = value

        apm = ElasticAPM(app=app, **apm_args)
        try:
            from sanic_testing import TestManager
        except ImportError:
            from sanic.testing import SanicTestClient as TestManager

        TestManager(app=app)

        bp = Blueprint(name="test", url_prefix="/apm", version="v1")

        @app.exception(ValueError)
        async def handle_value_error(request, exception):
            return json({"source": "value-error-custom"}, status=500)

        async def attribute_error_handler(request, expception):
            return json({"source": "custom-handler"}, status=500)

        app.error_handler.add(AttributeError, attribute_error_handler)

        @bp.post("/unhandled-exception")
        async def raise_something(request):
            raise CustomException("Unhandled")

        @app.route("/", methods=["GET", "POST"])
        def default_route(request: Request):
            with async_capture_span("test"):
                pass
            return json({"response": "ok"})

        @app.get("/greet/<name:str>")
        async def greet_person(request: Request, name: str):
            return json({"response": f"Hello {name}"})

        @app.get("/capture-exception")
        async def capture_exception(request):
            try:
                1 / 0
            except ZeroDivisionError:
                await apm.capture_exception()
            return json({"response": "invalid"}, 500)

        app.blueprint(blueprint=bp)

        @app.get("/raise-exception")
        async def raise_exception(request):
            raise AttributeError

        @app.get("/fallback-default-error")
        async def raise_default_error(request):
            raise CustomException

        @app.get("/raise-value-error")
        async def raise_value_error(request):
            raise ValueError

        @app.get("/add-custom-headers")
        async def custom_headers(request):
            return json({"data": "message"}, headers={"sessionid": 1234555})

        @app.get("/add-cookies")
        async def add_cookies(request):
            response = json({"data": "message"}, headers={"sessionid": 1234555})
            if hasattr(response, "add_cookie"):
                response.add_cookie("some", "cookie")
            else:
                response.cookies["some"] = "cookie"
            return response

        try:
            yield app, apm
        finally:
            elasticapm.uninstrument()

    return _generate



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("flask")  # isort:skip

import io
import logging
import os
from urllib.request import urlopen

import mock
from flask import signals

import elasticapm
from elasticapm.conf import constants
from elasticapm.conf.constants import ERROR, TRANSACTION
from elasticapm.contrib.flask import ElasticAPM
from elasticapm.utils.disttracing import TraceParent
from tests.contrib.flask.utils import captured_templates

pytestmark = pytest.mark.flask


def test_error_handler(flask_apm_client):
    client = flask_apm_client.app.test_client()
    response = client.get("/an-error/")
    assert response.status_code == 500
    assert len(flask_apm_client.client.events[ERROR]) == 1

    event = flask_apm_client.client.events[ERROR][0]

    assert "exception" in event
    exc = event["exception"]
    assert exc["type"] == "ValueError"
    assert exc["message"] == "ValueError: hello world"
    assert exc["handled"] is False
    assert event["culprit"] == "tests.contrib.flask.fixtures.an_error"

    transaction = flask_apm_client.client.events[TRANSACTION][0]
    assert transaction["result"] == "HTTP 5xx"
    assert transaction["name"] == "GET /an-error/"


def test_get(flask_apm_client):
    client = flask_apm_client.app.test_client()
    response = client.get("/an-error/?foo=bar")
    assert response.status_code == 500
    assert len(flask_apm_client.client.events[ERROR]) == 1

    event = flask_apm_client.client.events[ERROR][0]

    assert "request" in event["context"]
    request = event["context"]["request"]
    assert request["url"]["full"] == "http://localhost/an-error/?foo=bar"
    assert request["url"]["search"] == "?foo=bar"
    assert request["method"] == "GET"
    assert "body" not in request
    assert "headers" in request
    headers = request["headers"]
    assert "host" in headers, headers.keys()
    assert headers["host"] == "localhost"
    env = request["env"]
    assert "SERVER_NAME" in env, env.keys()
    assert env["SERVER_NAME"] == "localhost"
    assert "SERVER_PORT" in env, env.keys()
    assert env["SERVER_PORT"] == "80"


def test_get_debug(flask_apm_client):
    app = flask_apm_client.app
    app.debug = True
    app.config["TESTING"] = False
    with pytest.raises(ValueError):
        app.test_client().get("/an-error/?foo=bar")
    assert len(flask_apm_client.client.events) == 0


def test_get_debug_elasticapm(flask_apm_client):
    app = flask_apm_client.app
    app.debug = True
    app.config["TESTING"] = True
    flask_apm_client.client.config.debug = True
    with pytest.raises(ValueError):
        app.test_client().get("/an-error/?foo=bar")
    assert len(flask_apm_client.client.events[ERROR]) == 1


@pytest.mark.parametrize(
    "elasticapm_client", [{"capture_body": "errors"}, {"capture_body": "all"}, {"capture_body": "off"}], indirect=True
)
def test_post(flask_apm_client):
    client = flask_apm_client.app.test_client()
    response = client.post("/an-error/?biz=baz", data={"foo": "bar"})
    assert response.status_code == 500
    assert len(flask_apm_client.client.events[ERROR]) == 1

    event = flask_apm_client.client.events[ERROR][0]

    assert "request" in event["context"]
    request = event["context"]["request"]
    assert request["url"]["full"] == "http://localhost/an-error/?biz=baz"
    assert request["url"]["search"] == "?biz=baz"
    assert request["method"] == "POST"
    if flask_apm_client.client.config.capture_body in (constants.ERROR, "all"):
        assert request["body"] == {"foo": "bar"}
    else:
        assert request["body"] == "[REDACTED]"
    assert "headers" in request
    headers = request["headers"]
    assert "content-length" in headers, headers.keys()
    assert headers["content-length"] == "7"
    assert "content-type" in headers, headers.keys()
    assert headers["content-type"] == "application/x-www-form-urlencoded"
    assert "host" in headers, headers.keys()
    assert headers["host"] == "localhost"
    env = request["env"]
    assert "SERVER_NAME" in env, env.keys()
    assert env["SERVER_NAME"] == "localhost"
    assert "SERVER_PORT" in env, env.keys()
    assert env["SERVER_PORT"] == "80"


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"capture_body": "transactions"}, {"capture_body": "all"}, {"capture_body": "off"}],
    indirect=True,
)
def test_instrumentation(flask_apm_client):
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()

    assert resp.status_code == 200, resp.response

    transactions = flask_apm_client.client.events[TRANSACTION]

    assert len(transactions) == 1
    transaction = transactions[0]
    assert transaction["type"] == "request"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert "request" in transaction["context"]
    assert transaction["context"]["request"]["url"]["full"] == "http://localhost/users/"
    assert transaction["context"]["request"]["method"] == "POST"
    if flask_apm_client.client.config.capture_body in (constants.TRANSACTION, "all"):
        assert transaction["context"]["request"]["body"] == {"foo": "bar"}
    else:
        assert transaction["context"]["request"]["body"] == "[REDACTED]"
    assert transaction["context"]["response"]["status_code"] == 200
    assert transaction["context"]["response"]["headers"] == {
        "foo": "bar;baz",
        "bar": "bazzinga",
        "Content-Length": "78",
        "Content-Type": "text/html; charset=utf-8",
    }
    spans = flask_apm_client.client.spans_for_transaction(transactions[0])
    assert len(spans) == 1, [t["name"] for t in spans]

    expected_signatures = {"users.html"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "users.html"
    assert spans[0]["type"] == "template"
    assert spans[0]["subtype"] == "jinja2"
    assert spans[0]["action"] == "render"


def test_instrumentation_debug(flask_apm_client):
    flask_apm_client.app.debug = True
    assert len(flask_apm_client.client.events[TRANSACTION]) == 0
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()
    assert len(flask_apm_client.client.events[TRANSACTION]) == 0


@pytest.mark.parametrize("elasticapm_client", [{"debug": True}], indirect=True)
def test_instrumentation_debug_client_debug(flask_apm_client):
    flask_apm_client.app.debug = True
    assert len(flask_apm_client.client.events[TRANSACTION]) == 0
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()
    assert len(flask_apm_client.client.events[TRANSACTION]) == 1


def test_instrumentation_404(flask_apm_client):
    resp = flask_apm_client.app.test_client().post("/no-such-page/")
    resp.close()

    assert resp.status_code == 404, resp.response

    transactions = flask_apm_client.client.events[TRANSACTION]

    assert len(transactions) == 1
    spans = flask_apm_client.client.spans_for_transaction(transactions[0])
    assert transactions[0]["result"] == "HTTP 4xx"
    assert transactions[0]["context"]["response"]["status_code"] == 404
    assert len(spans) == 0, [t["signature"] for t in spans]


@pytest.mark.parametrize("header_name", [constants.TRACEPARENT_HEADER_NAME, constants.TRACEPARENT_LEGACY_HEADER_NAME])
def test_traceparent_handling(flask_apm_client, header_name):
    with mock.patch(
        "elasticapm.contrib.flask.TraceParent.from_string", wraps=TraceParent.from_string
    ) as wrapped_from_string:
        resp = flask_apm_client.app.test_client().post(
            "/users/",
            data={"foo": "bar"},
            headers={
                header_name: "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03",
                constants.TRACESTATE_HEADER_NAME: "foo=bar,baz=bazzinga",
            },
        )
        resp.close()

    assert resp.status_code == 200, resp.response

    transaction = flask_apm_client.client.events[TRANSACTION][0]

    assert transaction["trace_id"] == "0af7651916cd43dd8448eb211c80319c"
    assert transaction["parent_id"] == "b7ad6b7169203331"
    assert "foo=bar,baz=bazzinga" in wrapped_from_string.call_args[0]


def test_non_standard_http_status(flask_apm_client):
    resp = flask_apm_client.app.test_client().get("/non-standard-status/")
    resp.close()
    assert resp.status == "0 fail", resp.response
    assert resp.status_code == 0, resp.response

    transactions = flask_apm_client.client.events[TRANSACTION]
    assert transactions[0]["result"] == "0 fail"  # "0" is prepended by Werkzeug BaseResponse
    assert transactions[0]["context"]["response"]["status_code"] == 0


def test_framework_name(flask_app):
    apm = ElasticAPM(app=flask_app, metrics_interval="0ms")
    assert apm.client.config.framework_name == "flask"
    app_info = apm.client.get_service_info()
    assert app_info["framework"]["name"] == "flask"
    assert app_info["framework"]["version"]
    apm.client.close()

    # Cleanup -- we don't use the flask_apm_client fixture here because it uses
    # the elasticapm_client fixture, and so doesn't override the framework name.
    # However, that means we have to clean up manually
    elasticapm.uninstrument()
    signals.request_started.disconnect(apm.request_started)
    signals.request_finished.disconnect(apm.request_finished)
    # remove logging handler if it was added
    logger = logging.getLogger()
    for handler in list(logger.handlers):
        if getattr(handler, "client", None) is apm.client:
            logger.removeHandler(handler)


@pytest.mark.parametrize(
    "elasticapm_client", [{"capture_body": "errors"}, {"capture_body": "all"}, {"capture_body": "off"}], indirect=True
)
def test_post_files(flask_apm_client):
    with open(os.path.abspath(__file__), mode="rb") as f:
        response = flask_apm_client.app.test_client().post(
            "/an-error/",
            data={
                "foo": ["bar", "baz"],
                "f1": (io.BytesIO("1".encode("latin-1")), "bla"),
                "f2": [(f, "flask_tests.py"), (io.BytesIO("1".encode("latin-1")), "blub")],
            },
        )
    assert response.status_code == 500
    assert len(flask_apm_client.client.events[ERROR]) == 1

    event = flask_apm_client.client.events[ERROR][0]
    if flask_apm_client.client.config.capture_body in (constants.ERROR, "all"):
        assert event["context"]["request"]["body"] == {
            "foo": ["bar", "baz"],
            "_files": {"f1": "bla", "f2": ["flask_tests.py", "blub"]},
        }
    else:
        assert event["context"]["request"]["body"] == "[REDACTED]"


def test_capture_body_config_is_dynamic_for_errors(flask_apm_client):
    flask_apm_client.client.config.update(version="1", capture_body="all")
    resp = flask_apm_client.app.test_client().post("/an-error/", data={"foo": "bar"})
    resp.close()
    error = flask_apm_client.client.events[ERROR][0]
    assert error["context"]["request"]["body"] == {"foo": "bar"}

    flask_apm_client.client.config.update(version="2", capture_body="off")
    resp = flask_apm_client.app.test_client().post("/an-error/", data={"foo": "bar"})
    resp.close()
    error = flask_apm_client.client.events[ERROR][1]
    assert error["context"]["request"]["body"] == "[REDACTED]"


def test_capture_body_config_is_dynamic_for_transactions(flask_apm_client):
    flask_apm_client.client.config.update(version="1", capture_body="all")
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][0]
    assert transaction["context"]["request"]["body"] == {"foo": "bar"}

    flask_apm_client.client.config.update(version="2", capture_body="off")
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][1]
    assert transaction["context"]["request"]["body"] == "[REDACTED]"


def test_capture_headers_config_is_dynamic_for_errors(flask_apm_client):
    flask_apm_client.client.config.update(version="1", capture_headers=True)
    resp = flask_apm_client.app.test_client().post("/an-error/", data={"foo": "bar"})
    resp.close()
    error = flask_apm_client.client.events[ERROR][0]
    assert error["context"]["request"]["headers"]

    flask_apm_client.client.config.update(version="2", capture_headers=False)
    resp = flask_apm_client.app.test_client().post("/an-error/", data={"foo": "bar"})
    resp.close()
    error = flask_apm_client.client.events[ERROR][1]
    assert "headers" not in error["context"]["request"]


def test_capture_headers_config_is_dynamic_for_transactions(flask_apm_client):
    flask_apm_client.client.config.update(version="1", capture_headers=True)
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][0]
    assert transaction["context"]["request"]["headers"]

    flask_apm_client.client.config.update(version="2", capture_headers=False)
    resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][1]
    assert "headers" not in transaction["context"]["request"]


@pytest.mark.parametrize("elasticapm_client", [{"capture_body": "transactions"}], indirect=True)
def test_options_request(flask_apm_client):
    resp = flask_apm_client.app.test_client().options("/")
    resp.close()
    transactions = flask_apm_client.client.events[TRANSACTION]
    assert transactions[0]["context"]["request"]["method"] == "OPTIONS"


@pytest.mark.parametrize(
    "elasticapm_client", [{"capture_headers": "true"}, {"capture_headers": "false"}], indirect=True
)
def test_capture_headers_errors(flask_apm_client):
    resp = flask_apm_client.app.test_client().post("/an-error/", headers={"some-header": "foo"})
    resp.close()
    error = flask_apm_client.client.events[ERROR][0]
    if flask_apm_client.client.config.capture_headers:
        assert error["context"]["request"]["headers"]["some-header"] == "foo"
    else:
        assert "headers" not in error["context"]["request"]


@pytest.mark.parametrize(
    "elasticapm_client", [{"capture_headers": "true"}, {"capture_headers": "false"}], indirect=True
)
def test_capture_headers_transactions(flask_apm_client):
    resp = flask_apm_client.app.test_client().post("/users/", headers={"some-header": "foo"})
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][0]
    if flask_apm_client.client.config.capture_headers:
        assert transaction["context"]["request"]["headers"]["some-header"] == "foo"
        assert transaction["context"]["response"]["headers"]["foo"] == "bar;baz"
    else:
        assert "headers" not in transaction["context"]["request"]
        assert "headers" not in transaction["context"]["response"]


def test_streaming_response(flask_apm_client):
    resp = flask_apm_client.app.test_client().get("/streaming/")
    assert resp.data == b"01234"
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][0]
    spans = flask_apm_client.client.spans_for_transaction(transaction)
    assert transaction["duration"] > 50
    assert len(spans) == 5


def test_response_close_wsgi(flask_wsgi_server):
    # this tests the response-close behavior using a real WSGI server
    elasticapm_client = flask_wsgi_server.app.apm_client.client
    url = flask_wsgi_server.url + "/streaming/"
    response = urlopen(url)
    response.read()
    transaction = elasticapm_client.events[TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert transaction["duration"] > 50
    assert len(spans) == 5


def test_set_transaction_name(flask_apm_client):
    resp = flask_apm_client.app.test_client().get("/transaction-name/")
    resp.close()
    transaction = flask_apm_client.client.events[TRANSACTION][0]
    assert transaction["name"] == "foo"
    assert transaction["result"] == "okydoky"


def test_rum_tracing_context_processor(flask_apm_client):
    with captured_templates(flask_apm_client.app) as templates:
        resp = flask_apm_client.app.test_client().post("/users/", data={"foo": "bar"})
        resp.close()
        transaction = flask_apm_client.client.events[TRANSACTION][0]
        template, context = templates[0]
        assert context["apm"]["trace_id"] == transaction["trace_id"]
        assert context["apm"]["is_sampled"]
        assert context["apm"]["is_sampled_js"] == "true"
        assert callable(context["apm"]["span_id"])


@pytest.mark.parametrize("flask_apm_client", [{"logging": True}], indirect=True)
def test_logging_enabled(flask_apm_client):
    logger = logging.getLogger()
    logger.error("test")
    error = flask_apm_client.client.events[ERROR][0]
    assert error["log"]["level"] == "error"
    assert error["log"]["message"] == "test"


@pytest.mark.parametrize("flask_apm_client", [{"logging": False}], indirect=True)
def test_logging_disabled(flask_apm_client):
    logger = logging.getLogger()
    logger.error("test")
    assert len(flask_apm_client.client.events[ERROR]) == 0


@pytest.mark.parametrize("flask_apm_client", [{"logging": logging.ERROR}], indirect=True)
def test_logging_by_level(flask_apm_client):
    logger = logging.getLogger()
    logger.warning("test")
    logger.error("test")
    assert len(flask_apm_client.client.events[ERROR]) == 1
    error = flask_apm_client.client.events[ERROR][0]
    assert error["log"]["level"] == "error"


def test_flask_transaction_ignore_urls(flask_apm_client):
    resp = flask_apm_client.app.test_client().get("/users/")
    resp.close()
    assert len(flask_apm_client.client.events[TRANSACTION]) == 1

    flask_apm_client.client.config.update(version=1, transaction_ignore_urls="/user*")
    resp = flask_apm_client.app.test_client().get("/users/")
    resp.close()
    assert len(flask_apm_client.client.events[TRANSACTION]) == 1



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

httpx = pytest.importorskip("httpx")  # isort:skip
httpcore = pytest.importorskip("httpcore")  # isort:skip

import urllib.parse

from elasticapm.conf import constants
from elasticapm.conf.constants import TRANSACTION
from elasticapm.contrib.asyncio.traces import async_capture_span
from elasticapm.utils.disttracing import TraceParent

pytestmark = [pytest.mark.httpx, pytest.mark.asyncio]

httpcore_version = tuple(map(int, httpcore.__version__.split(".")[:3]))
httpx_version = tuple(map(int, httpx.__version__.split(".")[:3]))

if httpx_version < (0, 20):
    allow_redirects = {"allow_redirects": False}
else:
    allow_redirects = {"follow_redirects": False}


async def test_httpx_instrumentation(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction.test")
    async with async_capture_span("test_request", "test"):
        async with httpx.AsyncClient() as client:
            await client.get(url, **allow_redirects)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["outcome"] == "success"
    assert url == spans[0]["context"]["http"]["url"]
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    # this should be the span id of `httpx`, not of urllib3
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded


async def test_httpx_instrumentation_string_url(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    elasticapm_client.begin_transaction("transaction.test")
    async with async_capture_span("test_request", "test"):
        async with httpx.AsyncClient() as client:
            await client.get(url, **allow_redirects)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert url == spans[0]["context"]["http"]["url"]

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    # this should be the span id of `httpx`, not of urllib3
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded


async def test_httpx_instrumentation_malformed_empty(instrument, elasticapm_client):
    try:
        from httpx._exceptions import UnsupportedProtocol
    except ImportError:
        pytest.skip("Test requires HTTPX 0.14+")
    elasticapm_client.begin_transaction("transaction.test")
    async with async_capture_span("test_request", "test"):
        with pytest.raises(UnsupportedProtocol):
            httpx.get("")


async def test_httpx_instrumentation_malformed_path(instrument, elasticapm_client):
    try:
        from httpx._exceptions import LocalProtocolError, UnsupportedProtocol
    except ImportError:
        pytest.skip("Test requires HTTPX 0.14+")
    elasticapm_client.begin_transaction("transaction.test")
    async with async_capture_span("test_request", "test"):
        async with httpx.AsyncClient() as client:
            # raises UnsupportedProtocol since 0.18.0
            with pytest.raises((LocalProtocolError, UnsupportedProtocol)):
                await client.get("http://")


async def test_url_sanitization(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    url = url.replace("http://", "http://user:pass@")
    transaction_object = elasticapm_client.begin_transaction("transaction")
    async with httpx.AsyncClient() as client:
        await client.get(url)
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert "pass" not in span["context"]["http"]["url"]
    assert constants.MASK_URL in span["context"]["http"]["url"]


@pytest.mark.parametrize("status_code", [400, 500])
async def test_httpx_error(instrument, elasticapm_client, waiting_httpserver, status_code):
    waiting_httpserver.serve_content("", code=status_code)
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction")
    expected_sig = "GET {0}".format(parsed_url.netloc)
    url = "http://{0}/hello_world".format(parsed_url.netloc)
    async with async_capture_span("test_name", "test_type"):
        async with httpx.AsyncClient() as client:
            await client.get(url, **allow_redirects)

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert spans[0]["name"] == expected_sig
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["context"]["http"]["url"] == url
    assert spans[0]["context"]["http"]["status_code"] == status_code
    assert spans[0]["outcome"] == "failure"


async def test_httpx_streaming(instrument, elasticapm_client, waiting_httpserver):
    # client.stream passes the request as a keyword argument to the instrumented method.
    # This helps test that we can handle it both as an arg and a kwarg
    # see https://github.com/elastic/apm-agent-python/issues/1336
    elasticapm_client.begin_transaction("httpx-context-manager-client-streaming")
    client = httpx.AsyncClient()
    try:
        async with client.stream("GET", url=waiting_httpserver.url) as response:
            assert response
    finally:
        await client.aclose()
        elasticapm_client.end_transaction("httpx-context-manager-client-streaming")

    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]
    assert span["type"] == "external"
    assert span["subtype"] == "http"


@pytest.mark.skipif(httpcore_version < (0, 17, 3), reason="AsyncMockBackend not available on older versions")
async def test_default_port_handling(instrument, elasticapm_client):
    url = "https://example.com/"

    elasticapm_client.begin_transaction("transaction")
    network_backend = httpcore.AsyncMockBackend(
        buffer=[
            b"HTTP/1.1 200 OK\r\n",
            b"Content-Type: plain/text\r\n",
            b"Content-Length: 13\r\n",
            b"\r\n",
            b"Hello, world!",
        ]
    )
    async with httpcore.AsyncConnectionPool(network_backend=network_backend) as pool:
        await pool.request("GET", url)
    elasticapm_client.end_transaction("transaction")

    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert span["name"] == "GET example.com"
    assert span["context"]["http"]["url"] == url



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest

from elasticapm.instrumentation.packages.dbapi2 import (
    Literal,
    extract_action_from_signature,
    extract_signature,
    scan,
    tokenize,
)


def test_scan_simple():
    sql = "Hello 'Peter Pan' at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter Pan"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_with_escape_single_quote():
    sql = "Hello 'Peter\\' Pan' at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter' Pan"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_with_escape_slash():
    sql = "Hello 'Peter Pan\\\\' at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter Pan\\"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_double_quotes():
    sql = """Hello 'Peter'' Pan''' at Disney World"""
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("'", "Peter' Pan'"), "at", "Disney", "World"]
    assert actual == expected


def test_scan_double_quotes_at_end():
    sql = """Hello Peter Pan at Disney 'World'"""
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", "Peter", "Pan", "at", "Disney", Literal("'", "World")]
    assert actual == expected


@pytest.mark.parametrize("quote", ["$$", "$q$"])
@pytest.mark.parametrize(
    "content",
    [
        "",
        "q",
        "Peter q Pan",
        "Peter $ Pan",
        "Peter $q Pan",
        "Peter q$ Pan",
        "Peter $q q$ $q q$ Pan Peter $q q$ $q q$ Pan",
        "Peter $qq$ Pan",
    ],
)
def test_scan_dollar_quote(quote, content):
    sql = f"Hello {quote}{content}{quote} at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal(quote, content), "at", "Disney", "World"]
    assert actual == expected


def test_dollar_quote_containing_double_dollar():
    sql = "Hello $q$Peter $$ Pan$q$ at Disney World"
    tokens = tokenize(sql)
    actual = [t[1] for t in scan(tokens)]
    expected = ["Hello", Literal("$q$", "Peter $$ Pan"), "at", "Disney", "World"]
    assert actual == expected


def test_extract_signature_string():
    sql = "Hello 'Peter Pan' at Disney World"
    actual = extract_signature(sql)
    expected = "HELLO"
    assert actual == expected


def test_extract_signature_bytes():
    sql = b"Hello 'Peter Pan' at Disney World"
    actual = extract_signature(sql)
    expected = "HELLO"
    assert actual == expected


def test_extract_signature_pathological():
    # tune for performance testing
    multiplier = 10
    values = []
    for chunk in range(multiplier):
        i = chunk * 3
        values.append(f" (${1+i}::varchar, ${2+i}::varchar, ${3+i}::varchar), ")

    sql = f"SELECT * FROM (VALUES {''.join(values)})\n"
    actual = extract_signature(sql)
    expected = "SELECT FROM"
    assert actual == expected


@pytest.mark.parametrize(
    ["sql", "expected"],
    [
        (
            "EXEC AdventureWorks2022.dbo.uspGetEmployeeManagers 50;",
            "EXECUTE AdventureWorks2022.dbo.uspGetEmployeeManagers",
        ),
        ("EXECUTE sp_who2", "EXECUTE sp_who2"),
        ("EXEC sp_updatestats @@all_schemas = 'true'", "EXECUTE sp_updatestats"),
        ("CALL get_car_stats_by_year(2017, @number, @min, @avg, @max);", "CALL get_car_stats_by_year()"),
        ("CALL get_car_stats_by_year", "CALL get_car_stats_by_year()"),
        ("CALL get_car_stats_by_year;", "CALL get_car_stats_by_year()"),
        ("CALL get_car_stats_by_year();", "CALL get_car_stats_by_year()"),
    ],
)
def test_extract_signature_for_procedure_call(sql, expected):
    actual = extract_signature(sql)
    assert actual == expected


@pytest.mark.parametrize(
    ["sql", "expected"],
    [
        ("SELECT FROM table", "query"),
        ("EXEC sp_who", "exec"),
        ("EXECUTE sp_updatestats", "exec"),
        ("CALL me_maybe", "exec"),
    ],
)
def test_extract_action_from_signature(sql, expected):
    actual = extract_action_from_signature(sql, "query")
    assert actual == expected


@pytest.mark.parametrize(
    ["sql", "expected"],
    [
        ("SELECT username FROM user", "SELECT FROM user"),
        ("SELECT username FROM [user]", "SELECT FROM [user]"),
        ("SELECT username FROM [db].[user]", "SELECT FROM [db].[user]"),
        ("SELECT username FROM db.[user]", "SELECT FROM db.[user]"),
        ("SELECT username FROM [db].user", "SELECT FROM [db].user"),
    ],
)
def test_extract_signature_when_using_square_brackets(sql, expected):
    actual = extract_signature(sql)
    assert actual == expected



#  BSD 3-Clause License
#
#  Copyright (c) 2020, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

import itertools

import prometheus_client

from elasticapm.metrics.base_metrics import MetricSet


class PrometheusMetrics(MetricSet):
    def __init__(self, registry) -> None:
        super(PrometheusMetrics, self).__init__(registry)
        self._prometheus_registry = prometheus_client.REGISTRY

    def before_collect(self) -> None:
        for metric in self._prometheus_registry.collect():
            metric_type = self.METRIC_MAP.get(metric.type, None)
            if not metric_type:
                continue
            metric_type(self, metric.name, metric.samples, metric.unit)

    def _prom_counter_handler(self, name, samples, unit) -> None:
        # Counters can be converted 1:1 from Prometheus to our
        # format. Each pair of samples represents a distinct labelset for a
        # given name. The pair consists of the value, and a "created" timestamp.
        # We only use the former.
        for total_sample, _ in grouper(samples, 2):
            self.counter(self._registry.client.config.prometheus_metrics_prefix + name, **total_sample.labels).val = (
                total_sample.value
            )

    def _prom_gauge_handler(self, name, samples, unit) -> None:
        # Counters can be converted 1:1 from Prometheus to our
        # format. Each sample represents a distinct labelset for a
        # given name
        for sample in samples:
            self.gauge(self._registry.client.config.prometheus_metrics_prefix + name, **sample.labels).val = (
                sample.value
            )

    def _prom_summary_handler(self, name, samples, unit) -> None:
        # Prometheus Summaries are analogous to our Timers, having
        # a count and a sum. A prometheus summary is represented by
        # three values. The list of samples for a given name can be
        # grouped into 3-pairs of (count, sum, creation_timestamp).
        # Each 3-pair represents a labelset.
        for count_sample, sum_sample, _ in grouper(samples, 3):
            self.timer(self._registry.client.config.prometheus_metrics_prefix + name, **count_sample.labels).val = (
                sum_sample.value,
                count_sample.value,
            )

    def _prom_histogram_handler(self, name, samples, unit) -> None:
        # Prometheus histograms are structured as a series of counts
        # with an "le" label. The count of each label signifies all
        # observations with a lower-or-equal value with respect to
        # the "le" label value.
        # After the le-samples, 3 more samples follow, with an overall
        # count, overall sum, and creation timestamp.
        sample_pos = 0
        prev_val = 0
        counts = []
        values = []
        name = self._registry.client.config.prometheus_metrics_prefix + name
        while sample_pos < len(samples):
            sample = samples[sample_pos]
            if "le" in sample.labels:
                values.append(float(sample.labels["le"]))
                counts.append(int(sample.value - prev_val))
                prev_val = sample.value
                sample_pos += 1

            else:
                # we reached the end of one set of buckets/values, this is the "count" sample
                self.histogram(name, unit=unit, buckets=values, **sample.labels).val = counts
                prev_val = 0
                counts = []
                values = []
                sample_pos += 3  # skip sum/created samples

    METRIC_MAP = {
        "counter": _prom_counter_handler,
        "gauge": _prom_gauge_handler,
        "summary": _prom_summary_handler,
        "histogram": _prom_histogram_handler,
    }


def grouper(iterable, n, fillvalue=None):
    """Collect data into fixed-length chunks or blocks"""
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)



#  BSD 3-Clause License
#
#  Copyright (c) 2021, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import platform

import pytest

prometheus_client = pytest.importorskip("prometheus_client")

from elasticapm.metrics.base_metrics import MetricsRegistry
from elasticapm.metrics.sets.prometheus import PrometheusMetrics

pytestmark = pytest.mark.prometheus_client


prometheus_client.REGISTRY.unregister(prometheus_client.PROCESS_COLLECTOR)
prometheus_client.REGISTRY.unregister(prometheus_client.PLATFORM_COLLECTOR)
if platform.python_implementation() == "CPython":
    prometheus_client.REGISTRY.unregister(prometheus_client.GC_COLLECTOR)


@pytest.fixture()
def prometheus():
    # reset registry
    prometheus_client.REGISTRY._collector_to_names = {}
    prometheus_client.REGISTRY._names_to_collectors = {}


def test_counter(elasticapm_client, prometheus):
    metricset = PrometheusMetrics(MetricsRegistry(elasticapm_client))
    counter = prometheus_client.Counter("a_bare_counter", "Bare counter")
    counter_with_labels = prometheus_client.Counter(
        "counter_with_labels", "Counter with labels", ["alabel", "anotherlabel"]
    )
    counter.inc()
    counter_with_labels.labels(alabel="foo", anotherlabel="baz").inc()
    counter_with_labels.labels(alabel="bar", anotherlabel="bazzinga").inc()
    counter_with_labels.labels(alabel="foo", anotherlabel="baz").inc()
    data = list(metricset.collect())
    assert len(data) == 3
    assert data[0]["samples"]["prometheus.metrics.a_bare_counter"]["value"] == 1.0
    assert data[1]["samples"]["prometheus.metrics.counter_with_labels"]["value"] == 2.0
    assert data[1]["tags"] == {"alabel": "foo", "anotherlabel": "baz"}
    assert data[2]["samples"]["prometheus.metrics.counter_with_labels"]["value"] == 1.0
    assert data[2]["tags"] == {"alabel": "bar", "anotherlabel": "bazzinga"}


def test_gauge(elasticapm_client, prometheus):
    metricset = PrometheusMetrics(MetricsRegistry(elasticapm_client))
    gauge = prometheus_client.Gauge("a_bare_gauge", "Bare gauge")
    gauge_with_labels = prometheus_client.Gauge("gauge_with_labels", "Gauge with labels", ["alabel", "anotherlabel"])
    gauge.set(5)
    gauge_with_labels.labels(alabel="foo", anotherlabel="baz").set(7)
    gauge_with_labels.labels(alabel="bar", anotherlabel="bazzinga").set(11)
    gauge_with_labels.labels(alabel="foo", anotherlabel="baz").set(2)
    data = list(metricset.collect())
    assert len(data) == 3
    assert data[0]["samples"]["prometheus.metrics.a_bare_gauge"]["value"] == 5.0
    assert data[1]["samples"]["prometheus.metrics.gauge_with_labels"]["value"] == 2.0
    assert data[1]["tags"] == {"alabel": "foo", "anotherlabel": "baz"}
    assert data[2]["samples"]["prometheus.metrics.gauge_with_labels"]["value"] == 11.0
    assert data[2]["tags"] == {"alabel": "bar", "anotherlabel": "bazzinga"}


def test_summary(elasticapm_client, prometheus):
    metricset = PrometheusMetrics(MetricsRegistry(elasticapm_client))
    summary = prometheus_client.Summary("a_bare_summary", "Bare summary")
    summary_with_labels = prometheus_client.Summary(
        "summary_with_labels", "Summary with labels", ["alabel", "anotherlabel"]
    )
    summary.observe(5)
    summary.observe(7)
    summary.observe(9)
    summary_with_labels.labels(alabel="foo", anotherlabel="baz").observe(7)
    summary_with_labels.labels(alabel="bar", anotherlabel="bazzinga").observe(11)
    summary_with_labels.labels(alabel="foo", anotherlabel="baz").observe(2)
    data = list(metricset.collect())

    assert len(data) == 3
    assert data[0]["samples"]["prometheus.metrics.a_bare_summary.count"]["value"] == 3.0
    assert data[0]["samples"]["prometheus.metrics.a_bare_summary.sum"]["value"] == 21
    assert data[1]["samples"]["prometheus.metrics.summary_with_labels.count"]["value"] == 2.0
    assert data[1]["samples"]["prometheus.metrics.summary_with_labels.sum"]["value"] == 9.0
    assert data[1]["tags"] == {"alabel": "foo", "anotherlabel": "baz"}
    assert data[2]["samples"]["prometheus.metrics.summary_with_labels.count"]["value"] == 1.0
    assert data[2]["samples"]["prometheus.metrics.summary_with_labels.sum"]["value"] == 11.0
    assert data[2]["tags"] == {"alabel": "bar", "anotherlabel": "bazzinga"}


def test_histogram(elasticapm_client, prometheus):
    metricset = PrometheusMetrics(MetricsRegistry(elasticapm_client))
    histo = prometheus_client.Histogram("histo", "test histogram", buckets=[1, 10, 100, float("inf")])
    histo_with_labels = prometheus_client.Histogram(
        "histowithlabel", "test histogram with labels", ["alabel", "anotherlabel"], buckets=[1, 10, 100, float("inf")]
    )
    histo.observe(0.5)
    histo.observe(0.6)
    histo.observe(1.5)
    histo.observe(26)
    histo.observe(42)
    histo.observe(12)
    histo.observe(105)
    histo_with_labels.labels(alabel="foo", anotherlabel="baz").observe(1)
    histo_with_labels.labels(alabel="foo", anotherlabel="baz").observe(10)
    histo_with_labels.labels(alabel="foo", anotherlabel="baz").observe(100)
    histo_with_labels.labels(alabel="foo", anotherlabel="bazzinga").observe(1000)
    data = list(metricset.collect())
    assert data[0]["samples"]["prometheus.metrics.histo"]["values"] == [0.5, 5.5, 55.0, 100.0]
    assert data[0]["samples"]["prometheus.metrics.histo"]["counts"] == [2, 1, 3, 1]
    assert all(isinstance(v, int) for v in data[0]["samples"]["prometheus.metrics.histo"]["counts"])

    assert data[1]["samples"]["prometheus.metrics.histowithlabel"]["values"] == [0.5, 5.5, 55.0, 100.0]
    assert data[1]["samples"]["prometheus.metrics.histowithlabel"]["counts"] == [1, 1, 1, 0]
    assert all(isinstance(v, int) for v in data[1]["samples"]["prometheus.metrics.histowithlabel"]["counts"])
    assert data[1]["tags"] == {"alabel": "foo", "anotherlabel": "baz"}

    assert data[2]["samples"]["prometheus.metrics.histowithlabel"]["values"] == [0.5, 5.5, 55.0, 100.0]
    assert data[2]["samples"]["prometheus.metrics.histowithlabel"]["counts"] == [0, 0, 0, 1]
    assert all(isinstance(v, int) for v in data[2]["samples"]["prometheus.metrics.histowithlabel"]["counts"])
    assert data[2]["tags"] == {"alabel": "foo", "anotherlabel": "bazzinga"}



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import mock
import pytest

import elasticapm


def test_bare_transaction(elasticapm_client):
    elasticapm_client.begin_transaction("request", start=0)
    elasticapm_client.end_transaction("test", "OK", duration=5)
    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    data = list(breakdown.collect())
    assert len(data) == 1
    asserts = 0
    for elem in data:
        if "span.self_time.sum.us" in elem["samples"]:
            assert elem["samples"]["span.self_time.count"]["value"] == 1
            assert elem["samples"]["span.self_time.sum.us"]["value"] == 5000000
            assert elem["transaction"] == {"name": "test", "type": "request"}
            assert elem["span"] == {"subtype": "", "type": "app"}
            asserts += 1
    assert asserts == 1


def test_single_span(elasticapm_client):
    elasticapm_client.begin_transaction("request", start=0)
    with elasticapm.capture_span("test", span_type="db", span_subtype="mysql", start=10, duration=5):
        pass
    elasticapm_client.end_transaction("test", "OK", duration=15)
    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    data = list(breakdown.collect())
    assert len(data) == 2
    asserts = 0
    for elem in data:
        if "span.self_time.sum.us" in elem["samples"]:
            if elem["span"] == {"type": "app", "subtype": ""}:
                assert elem["transaction"] == {"name": "test", "type": "request"}
                assert elem["samples"]["span.self_time.sum.us"]["value"] == 10000000
                assert elem["samples"]["span.self_time.count"]["value"] == 1
                asserts += 1
            elif elem["span"] == {"type": "db", "subtype": "mysql"}:
                assert elem["samples"]["span.self_time.count"]["value"] == 1
                assert elem["samples"]["span.self_time.sum.us"]["value"] == 5000000
                assert elem["transaction"] == {"name": "test", "type": "request"}
                asserts += 1
    assert asserts == 2


def test_nested_spans(elasticapm_client):
    elasticapm_client.begin_transaction("request", start=0)
    with elasticapm.capture_span("test", span_type="template", span_subtype="django", start=5, duration=15):
        with elasticapm.capture_span("test", span_type="db", span_subtype="mysql", start=10, duration=5):
            pass
        with elasticapm.capture_span("test", span_type="db", span_subtype="mysql", start=15, duration=5):
            pass
    elasticapm_client.end_transaction("test", "OK", duration=25)
    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    data = list(breakdown.collect())
    assert len(data) == 3
    asserts = 0
    for elem in data:
        if "span.self_time.sum.us" in elem["samples"]:
            if elem["span"] == {"type": "app", "subtype": ""}:
                assert elem["transaction"] == {"name": "test", "type": "request"}
                assert elem["samples"]["span.self_time.sum.us"]["value"] == 10000000
                assert elem["samples"]["span.self_time.count"]["value"] == 1
                asserts += 1
            elif elem["span"] == {"type": "db", "subtype": "mysql"}:
                assert elem["samples"]["span.self_time.count"]["value"] == 2
                assert elem["samples"]["span.self_time.sum.us"]["value"] == 10000000
                assert elem["transaction"] == {"name": "test", "type": "request"}
                asserts += 1
            elif elem["span"] == {"type": "template", "subtype": "django"}:
                assert elem["samples"]["span.self_time.count"]["value"] == 1
                assert elem["samples"]["span.self_time.sum.us"]["value"] == 5000000
                assert elem["transaction"] == {"name": "test", "type": "request"}
                asserts += 1
    assert asserts == 3


def test_explicit_app_span(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("request")
    with elasticapm.capture_span("test", span_type="app"):
        pass
    elasticapm_client.end_transaction("test", "OK")
    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    data = list(breakdown.collect())
    assert len(data) == 1
    asserts = 0
    for elem in data:
        if "span.self_time.sum.us" in elem["samples"]:
            if elem["span"] == {"type": "app", "subtype": ""}:
                assert elem["transaction"] == {"name": "test", "type": "request"}
                assert elem["samples"]["span.self_time.count"]["value"] == 2
                asserts += 1
    assert asserts == 1


@pytest.mark.parametrize("elasticapm_client", [{"breakdown_metrics": False}], indirect=True)
def test_disable_breakdowns(elasticapm_client):
    with pytest.raises(LookupError):
        elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    with mock.patch("elasticapm.traces.BaseSpan.child_started") as mock_child_started, mock.patch(
        "elasticapm.traces.Transaction.track_span_duration"
    ) as mock_track_span_duration:
        transaction = elasticapm_client.begin_transaction("test")
        assert transaction._breakdown is None
        with elasticapm.capture_span("test", span_type="template", span_subtype="django", duration=5):
            pass
        elasticapm_client.end_transaction("test", "OK", duration=5)
        assert mock_child_started.call_count == 0
        assert mock_track_span_duration.call_count == 0


def test_metrics_reset_after_collect(elasticapm_client):
    elasticapm_client.begin_transaction("request")
    with elasticapm.capture_span("test", span_type="db", span_subtype="mysql", duration=5):
        pass
    elasticapm_client.end_transaction("test", "OK", duration=15)
    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    for labels, c in breakdown._counters.items():
        assert c.val != 0
    for labels, t in breakdown._timers.items():
        assert t.val != (0, 0)
    list(breakdown.collect())
    for labels, c in breakdown._counters.items():
        assert c.val == 0
    for labels, t in breakdown._timers.items():
        assert t.val == (0, 0)


def test_multiple_transactions(elasticapm_client):
    for i in (1, 2):
        elasticapm_client.begin_transaction("request")
        with elasticapm.capture_span("test", duration=5):
            pass
        elasticapm_client.end_transaction("test", "OK", duration=10)

    breakdown = elasticapm_client.metrics.get_metricset("elasticapm.metrics.sets.breakdown.BreakdownMetricSet")
    data = list(breakdown.collect())
    asserts = 0
    for elem in data:
        if "span.self_time.sum.us" in elem["samples"]:
            if elem["span"] == {"type": "app", "subtype": ""}:
                assert elem["transaction"] == {"name": "test", "type": "request"}
                # precision lost due to float arithmetic
                assert 9999999 <= elem["samples"]["span.self_time.sum.us"]["value"] <= 10000000
                assert elem["samples"]["span.self_time.count"]["value"] == 2
                asserts += 1
            elif elem["span"] == {"type": "code", "subtype": "custom"}:
                assert elem["transaction"] == {"name": "test", "type": "request"}
                assert 9999999 <= elem["samples"]["span.self_time.sum.us"]["value"] <= 10000000
                assert elem["samples"]["span.self_time.count"]["value"] == 2
                asserts += 1
    assert asserts == 2
